local ADDON_NAME	= "ControlOutline"
local CODE_VERSION 	= 1.9

--===================================--
--===================================--
local SAVED_VAR_VERSION = 1.6
--===================================--
--===================================--
	
ControlOutline = {
	pool = {},
	outlinedControls = {},
	sv = {},
}

local outlinedControls = ControlOutline.outlinedControls
local outlinesPool = ControlOutline.pool


local function GetNextEdgeColorKey(edgeColorKey)
	local numColors = ControlOutline.sv["NUM_EDGE_COLORS"]
	
	--local newKey = (edgeColorKey % numColors) + 1
	return (edgeColorKey % numColors) + 1
end

--===================================================--
--======== Create Control Outline =========--
--===================================================--
function ControlOutline.CreateOutline(control, edgeColorKey, startingParentCtrl)
	-- Does an outline already exist
	local controlKey = control.ControlOutlinePoolKey
	local controlOutlineStartingParentControl =	control.ControlOutlineStartingParentControl
	if controlKey or controlOutlineStartingParentControl then return false end
	
	local outlineFrameCtrl, iKey = outlinesPool:AcquireObject()
	local outlineControl         = outlineFrameCtrl:GetNamedChild("Outline")
	
	-- Save the key in the control & save a reference to the control
	-- for clearing it out later
	control.ControlOutlinePoolKey = iKey
	-->Child control will get the startingParentControl as an attribut so that
	-->outlines of a parent will remove all the children that were outlined with it before
	if startingParentCtrl ~= nil and control ~= startingParentCtrl then
		control.ControlOutlineStartingParentControl = startingParentCtrl
	end
	outlinedControls[iKey] = control
	
	outlineFrameCtrl:ClearAnchors()
	outlineFrameCtrl:SetAnchor(TOPLEFT, control, TOPLEFT)
	outlineFrameCtrl:SetAnchor(BOTTOMRIGHT, control, BOTTOMRIGHT)
	
	if ControlOutline.sv["CHILD_OUTLINES"] then
		outlineFrameCtrl:SetParent(control)
	end
	
	edgeColorKey = edgeColorKey or 1
	local edgeColorTable = ControlOutline.sv["EDGE_COLORS"][edgeColorKey]
	
	outlineControl:SetEdgeColor(unpack(edgeColorTable))
	outlineControl:SetEdgeTexture(nil, 2, 2, ControlOutline.sv["EDGE_SIZE"], 0)
end
local createOutline = ControlOutline.CreateOutline

--===================================================--
--======== GetChildren() =========--
--===================================================--
function ControlOutline.GetChildren(control)
	if not control.GetChildren then
		control.GetChildren = control.GetChildren or function(self)
			local children = {}
			local numChildren = self:GetNumChildren()
			
			for childIndex=1, numChildren do
				local child = self:GetChild(childIndex)
				if child then
					children[childIndex] = child
				end
			end
			return children
		end
	end
	
	return control:GetChildren()
end
local getChildren = ControlOutline.GetChildren

local function releasePoolObjectAndHideOutline(control)
	local controlKey = control.ControlOutlinePoolKey
	if controlKey then
		outlinesPool:ReleaseObject(controlKey)
		control.ControlOutlinePoolKey = nil
		control.ControlOutlineStartingParentControl = nil
		outlinedControls[controlKey] = nil
		return true
	end
	return false
end

--===================================================--
--======== Outline Controls =========--
--===================================================--

function ControlOutline_ToggleOutline(control)
	control = control or moc()
	if control == GuiRoot then return end

	if releasePoolObjectAndHideOutline(control) then
		return
	end

	-- Outline parent control
	createOutline(control, 1)
end


local outlineControlAndChildren
function ControlOutline.OutlineControlAndChildren(control, edgeColorKey, startingParentCtrl)
	outlineControlAndChildren = outlineControlAndChildren or ControlOutline.OutlineControlAndChildren
	--Do not outline again if already outlined control
	local controlKey = control.ControlOutlinePoolKey
	if controlKey ~= nil then
		return
	end

	edgeColorKey = edgeColorKey or 1
	local children = getChildren(control)

	-- Outline Children first
	for _,childControl in pairs(children) do
		local nextEdgeColorKey = GetNextEdgeColorKey(edgeColorKey)
		--Get children of children
		outlineControlAndChildren(childControl, nextEdgeColorKey, startingParentCtrl)
	end

	-- Outline parent/child control
	if ControlOutline.sv["OUTLINE_HIDDEN_CONTROLS"] or not control:IsHidden() then
		createOutline(control, edgeColorKey, startingParentCtrl)
	end
end
outlineControlAndChildren = ControlOutline.OutlineControlAndChildren

--===================================================--
--======== Starting moc() Function =========--
--===================================================--
function ControlOutline_OutlineParentChildControls(control)
	control = control or moc()
	if control == GuiRoot then return end
	
	outlineControlAndChildren(control, nil, control)
end

--===================================================--
--======== Release =========--
--===================================================--
function ControlOutline_ReleaseOutlines(control)
	control = control or moc()
	if control == GuiRoot then return end

	for _, controlOfControl in pairs(outlinedControls) do
		if controlOfControl.ControlOutlineStartingParentControl and controlOfControl.ControlOutlineStartingParentControl == control then
			releasePoolObjectAndHideOutline(controlOfControl)
		end
	end
	releasePoolObjectAndHideOutline(control)
end

function ControlOutline_ReleaseAllOutlines()
	outlinesPool:ReleaseAllObjects()
	
	for _, control in pairs(outlinedControls) do
		control.ControlOutlinePoolKey = nil
	end
	outlinedControls = {}
end


--===================================================--
--======== Check and Is functions =========--
--===================================================--
function ControlOutline_IsControlOutlined(control)
	local key = ZO_KeyOfFirstElementInNonContiguousTable(outlinedControls, control)
	return (key ~= nil and true) or false, key
end


-------------------------------------------------------------------
--  On AddOn loaded  --
-------------------------------------------------------------------
function ControlOutline.Initialize()
--d("ControlOutline.Initialize")
	--===== SI Bindings =====--
	ZO_CreateStringId("SI_BINDING_NAME_CONTROL_OUTLINE_TOGGLE_OUTLINE", "Toggle outline on control (below mouse)")
	ZO_CreateStringId("SI_BINDING_NAME_CONTROL_OUTLINE_PARENT_N_CHILDREN_CONTROLS", "Outline control & children (below mouse)")
	ZO_CreateStringId("SI_BINDING_NAME_CONTROL_OUTLINE_RELEASE_PARENT_OUTLINES", "Hide control & children's outlines (below mouse)")
	ZO_CreateStringId("SI_BINDING_NAME_CONTROL_OUTLINE_RELEASE_ALL_OUTLINES", "Hide all outlines")

	--===== Control Pool =====--
	ControlOutline.pool = ZO_ControlPool:New("ControlOutline_DevOutline", nil)
	outlinesPool = ControlOutline.pool
	local function CustomPoolResetBehavior(control)
		control:SetParent(nil)
	end
	outlinesPool:SetCustomResetBehavior(CustomPoolResetBehavior)
	
	--===== Saved Vars =====--
	local defaultSavedVars = {
		["OUTLINE_HIDDEN_CONTROLS"] = false,
		["CHILD_OUTLINES"] 			= true,
		["EDGE_SIZE"] 				= 4,
		["NUM_EDGE_COLORS"] 		= 3,
		["EDGE_COLORS"] = {
			[1] = {1, 0, 0, 1},	-- Red
			[2] = {0, 1, 0, 1},	-- Green
			[3] = {0, 0, 1, 1},	-- Blue
			[4] = {1, 1, 0, 1},	-- Yellow
			[5] = {0, 1, 1, 1},	-- Cyan
			[6] = {1, 1, 1, 1},	-- White
		},
	}
	ControlOutline.sv = ZO_SavedVars:NewAccountWide(ADDON_NAME .."SavedVars", SAVED_VAR_VERSION, GetWorldName(), defaultSavedVars)
	
	--==== Settings Menu ====--
	ControlOutline.CreateSettingsMenu()
end

--==============================--
--======  Settings Menu =======--
--==============================--
function ControlOutline.CreateSettingsMenu()
--d("ControlOutline.CreateSettingsMenu")

	local LAM2 = LibAddonMenu2

	local panelData = {
		type = "panel",
		name = ADDON_NAME,
		displayName = "Control Outlines",
		author = "Circonian & Baertram",
		version = tostring(CODE_VERSION),
		registerForRefresh = true,
		registerForDefaults = true,
	}
	local cntrlOptionsPanel = LAM2:RegisterAddonPanel(ADDON_NAME .. "_LAM2_Panel", panelData)

	local settings = ControlOutline.sv

	local optionsData = {
		{
			type = "description",
			text = "Open the control settings to specify the keybindings to add/remove the outlines.",
		},
		{
			type = "checkbox",
			name = "Outline Hidden Controls",
			tooltip = "When OFF it will only outline controls that are not hidden. When ON it will outline all controls whether they are visible or not (can slow things down).",
			default = false,
			getFunc = function() return settings["OUTLINE_HIDDEN_CONTROLS"] end,
			setFunc = function(bValue) settings["OUTLINE_HIDDEN_CONTROLS"] = bValue end,
		},
		{
			type = "checkbox",
			name = "Child Outlines",
			tooltip = "When ON the outlines will become children of the control they are outlining. This will allow them to be shown/hidden with the parent, BUT on some game controls it can cause errors to occur.",
			default = false,
			getFunc = function() return settings["CHILD_OUTLINES"] end,
			setFunc = function(bValue) settings["CHILD_OUTLINES"] = bValue end,
		},
		{
			type = "slider",
			name = "Edge Size",
			tooltip = "Adjust the size/thickness of the outline.",
			min = 0.5,
			max = 10,
			step = 0.5,
			decimals = 1,
			clampInput = true,
			default = 3,
			getFunc = function() return settings["EDGE_SIZE"] end,
			setFunc = function(iValue) settings["EDGE_SIZE"] = iValue end,
		},
		{
			type = "slider",
			name = "Number Of Edge Colors",
			tooltip = "When greater than 1: the edge colors will alternate using this many colors. The parent will use the 1st edge color, its children will use the 2nd edge color, those childrens children will use the 3rd edge color, exc...",
			min = 1,
			max = 6,
			step = 1,
			default = 1,
			getFunc = function() return settings["NUM_EDGE_COLORS"] end,
			setFunc = function(iValue) settings["NUM_EDGE_COLORS"] = iValue end,
		},
		{
			type = "submenu",
			name = "Edge Colors",
			controls = {
				{
					type = "colorpicker",
					name = "First Edge Color",
					tooltip = "Changes the first level outline color.",
					getFunc = function() return unpack(settings["EDGE_COLORS"][1]) end,
					setFunc = function(r,g,b,a) settings["EDGE_COLORS"][1] = {r,g,b,a} end,
				},
				{
					type = "colorpicker",
					name = "Second Edge Color",
					tooltip = "Changes the second level outline color.",
					getFunc = function() return unpack(settings["EDGE_COLORS"][2]) end,
					setFunc = function(r,g,b,a) settings["EDGE_COLORS"][2] = {r,g,b,a} end,
				},
				{
					type = "colorpicker",
					name = "Third Edge Color",
					tooltip = "Changes the third level outline color",
					getFunc = function() return unpack(settings["EDGE_COLORS"][3]) end,
					setFunc = function(r,g,b,a) settings["EDGE_COLORS"][3] = {r,g,b,a} end,
				},
				{
					type = "colorpicker",
					name = "Fourth Edge Color",
					tooltip = "Changes the third level outline color.",
					getFunc = function() return unpack(settings["EDGE_COLORS"][4]) end,
					setFunc = function(r,g,b,a) settings["EDGE_COLORS"][4] = {r,g,b,a} end,
				},
				{
					type = "colorpicker",
					name = "Fifth Edge Color",
					tooltip = "Changes the fifth level outline color.",
					getFunc = function() return unpack(settings["EDGE_COLORS"][5]) end,
					setFunc = function(r,g,b,a) settings["EDGE_COLORS"][5] = {r,g,b,a} end,
				},
				{
					type = "colorpicker",
					name = "Sixth Edge Color",
					tooltip = "Changes the sixth level outline color.",
					getFunc = function() return unpack(settings["EDGE_COLORS"][6]) end,
					setFunc = function(r,g,b,a) settings["EDGE_COLORS"][6] = {r,g,b,a} end,
				},
			},
		},
	}
	LAM2:RegisterOptionControls(ADDON_NAME .. "_LAM2_Panel", optionsData)
end


-------------------------------------------------------------------
--  OnAddOnLoaded  --
-------------------------------------------------------------------
local function OnAddOnLoaded(_, addonName)
	if addonName ~= ADDON_NAME then return end
	EVENT_MANAGER:UnregisterForEvent(ADDON_NAME, EVENT_ADD_ON_LOADED)

	ControlOutline.Initialize()
end

---------------------------------------------------------------------
--  Register Events --
---------------------------------------------------------------------
EVENT_MANAGER:RegisterForEvent(ADDON_NAME, EVENT_ADD_ON_LOADED, OnAddOnLoaded)